local sdk = require "sdk.sdk"
local drawManager = require "drawManager"
local player = require "player"

local Orc = {}
Orc.__index = Orc

function loadAnimation(imagePath, frameCount, directionCount)
    local image = love.graphics.newImage(imagePath)
    local frames = {}
    local imageWidth = image:getWidth()
    local imageHeight = image:getHeight()

    for i = 0, directionCount - 1 do
        frames[i] = {}
        for j = 0, frameCount - 1 do
            frames[i][j + 1] = love.graphics.newQuad(
                                   j * imageWidth / frameCount,
                                   i * imageHeight / directionCount,
                                   imageWidth / frameCount,
                                   imageHeight / directionCount, imageWidth,
                                   imageHeight)
        end
    end

    return {
        image = image,
        frames = frames,
        frameCount = frameCount,
        directionCount = directionCount
    }
end

local animations = {
    idle = loadAnimation('img/Orc/Idle.png', 16, 4),
    walk = loadAnimation('img/Orc/Walk.png', 4, 4),
    attack = loadAnimation('img/Orc/Attack.png', 4, 4),
    damage = loadAnimation('img/Orc/Dmg.png', 4, 4),
    die = loadAnimation('img/Orc/Die.png', 12, 1)
}

function Orc.new(id, x, y, data)
    local self = setmetatable({}, Orc)
    self.id = id
    self.x = x
    self.y = y
    self.scale = 6
    self.speed = 200
    self.direction = data.direction or 's'
    self.facing = data.facing or 's'
    self.currentFrame = 1
    self.animationTime = 0
    self.attackDuration = 0.8
    self.state = data.state or 'idle'
    self.spriteBottomBuffer = 13 -- annoying buffer of transparent pixels
    self.targetX = x
    self.targetY = y
    self.lastUpdate = 0
    self.type = "orc"
    self.animations = animations
    self.knockback = false
    self.knockbackDuration = 0.5
    self.knockbackStartTime = 0
    self.knockbackSpeed = 300
    self.knockbackX = 0
    self.knockbackY = 0
    self.lastKillMessageTime = 0
    return self
end

function Orc:load() end

function Orc:update(dt, x, y, data)
    self.targetX = x
    self.targetY = y

    -- resend kill message (handles it getting lost)
    if self.state == 'damage' then
        local currentTime = love.timer.getTime()
        if currentTime - self.lastKillMessageTime >= 0.2 then
            sdk.message({kill = true, entityId = self.id})
            self.lastKillMessageTime = currentTime
        end
    end

    if self.state ~= 'damage' and self.state ~= 'die' then
        if data.state ~= self.state then self.currentFrame = 1 end
        if self.lastUpdate ~= data.lastUpdate then
            self.lastUpdate = data.lastUpdate
            self.interpolationStartTime = love.timer.getTime()
            self.startX = self.x
            self.startY = self.y
        end
        local elapsedTime = love.timer.getTime() - self.interpolationStartTime
        if elapsedTime < 0.3 then
            local lerpFactor = elapsedTime / 0.3
            self.x = self.startX + (self.targetX - self.startX) * lerpFactor
            self.y = self.startY + (self.targetY - self.startY) * lerpFactor
        else
            self.x = self.targetX
            self.y = self.targetY
        end
        self.state = data.state or 'idle'
    else
        local newState =
            (self.state == 'die' or data.state == 'die') and 'die' or 'damage'
        if newState ~= self.state then self.currentFrame = 1 end
        self.state = newState
        if self.knockback then
            local elapsedTime = love.timer.getTime() - self.knockbackStartTime
            if elapsedTime < self.knockbackDuration then
                local t = elapsedTime / self.knockbackDuration
                local speedFactor = (1 - t * t) * self.knockbackSpeed
                self.x = self.x + self.knockbackX * speedFactor * dt
                self.y = self.y + self.knockbackY * speedFactor * dt
            else
                self.knockback = false
            end
        end
    end

    if data.state == 'die' then
        self.direction = 's'
        self.facing = 's'
    else
        self.direction = data.direction or 's'
        self.facing = data.facing or 's'
    end
    self.animationTime = self.animationTime + dt
    if self.animationTime > 0.2 then
        if self.state == 'attack' and self.currentFrame ==
            self.animations.attack.frameCount then self.state = 'idle' end
        local animation = self.animations[self.state]
        if self.state == 'die' then
            self.currentFrame = self.currentFrame + 1
            if self.currentFrame > animation.frameCount then
                self.currentFrame = animation.frameCount
            end
        else
            self.currentFrame = self.currentFrame % animation.frameCount + 1
        end
        self.animationTime = self.animationTime - 0.2
    end
end

function Orc:draw()
    local directionIndex
    if self.state == 'die' then
        directionIndex = 0
    elseif self.direction == 's' then
        directionIndex = 0
    elseif self.direction == 'n' then
        directionIndex = 2
    elseif self.direction == 'e' then
        directionIndex = self.facing == 's' and 0 or 2
    elseif self.direction == 'w' then
        directionIndex = self.facing == 's' and 1 or 3
    end

    local animation = self.animations[self.state]
    local frame = animation.frames[directionIndex][self.currentFrame %
                      animation.frameCount + 1]
    local frameWidth, frameHeight = animation.image:getWidth() /
                                        animation.frameCount,
                                    animation.image:getHeight() /
                                        animation.directionCount
    local scaledWidth, scaledHeight = frameWidth * self.scale,
                                      frameHeight * self.scale
    local offsetX, offsetY = scaledWidth / 2, scaledHeight / 2
    local drawX = self.x - offsetX
    local drawY = self.y - offsetY
    drawManager.bufferDrawEntity(drawX, drawY + (2 * offsetY) -
                                     (self.spriteBottomBuffer * self.scale),
                                 function()
        love.graphics.draw(animation.image, frame, drawX, drawY, 0, self.scale,
                           self.scale)
    end)
end

function Orc:getBoundingBox()
    local frameWidth, frameHeight = 10, 10
    local scaledWidth, scaledHeight = frameWidth * self.scale,
                                      frameHeight * self.scale

    return self.x - scaledWidth / 2, self.y - scaledHeight / 2, scaledWidth,
           scaledHeight
end

function Orc:onCollision()
    if self.state ~= 'damage' and self.state ~= 'die' then
        self.state = 'damage'
        self.lastKillMessageTime = love.timer.getTime()
        self.currentFrame = 1
        sdk.message({kill = true, entityId = self.id})
        -- knockback
        local dx = self.x - player.x
        local dy = self.y - player.y
        local mag = math.sqrt(dx * dx + dy * dy)
        self.knockbackX = dx / mag
        self.knockbackY = dy / mag
        self.knockback = true
        self.knockbackStartTime = love.timer.getTime()
    end
end

return Orc
