local sdk = require "sdk.sdk"
local drawManager = require "drawManager"
local pixelFont = love.graphics.newFont("font/PixelAE-Regular.ttf", 16)

local OtherPlayer = {}
OtherPlayer.__index = OtherPlayer

function loadAnimation(imagePath, frameCount, directionCount)
    local image = love.graphics.newImage(imagePath)
    local frames = {}
    local imageWidth = image:getWidth()
    local imageHeight = image:getHeight()

    for i = 0, directionCount - 1 do
        frames[i] = {}
        for j = 0, frameCount - 1 do
            frames[i][j + 1] = love.graphics.newQuad(
                                   j * imageWidth / frameCount,
                                   i * imageHeight / directionCount,
                                   imageWidth / frameCount,
                                   imageHeight / directionCount, imageWidth,
                                   imageHeight)
        end
    end

    return {
        image = image,
        frames = frames,
        frameCount = frameCount,
        directionCount = directionCount
    }
end

local animations = {
    idle = loadAnimation('img/Human/Idle.png', 16, 4),
    walk = loadAnimation('img/Human/Walk.png', 4, 4),
    attack = loadAnimation('img/Human/Attack.png', 4, 4),
    damage = loadAnimation('img/Human/Dmg.png', 4, 4),
    die = loadAnimation('img/Human/SoulDie.png', 12, 1)
}

function OtherPlayer.new(id, x, y, data)
    local self = setmetatable({}, OtherPlayer)
    self.id = id
    self.x = x
    self.y = y
    self.scale = 6
    self.speed = 200
    self.direction = data.direction or 's'
    self.facing = data.facing or 's'
    self.currentFrame = 1
    self.animationTime = 0
    self.attackDuration = 0.8
    self.state = data.state or 'idle'
    self.spriteBottomBuffer = 13 -- annoying buffer of transparent pixels
    self.targetX = x
    self.targetY = y
    self.lastUpdate = 0
    self.health = data.health
    self.type = "player"
    self.animations = animations
    self.displayName = data.displayName
    self.isFlashing = false
    self.flashTimer = 0
    self.isVisible = true
    return self
end

function OtherPlayer:load() end

function OtherPlayer:update(dt, x, y, data)
    self.targetX = x
    self.targetY = y
    self.displayName = data.displayName
    self.isFlashing = data.isFlashing or false

    if self.lastUpdate ~= data.lastUpdate then
        self.lastUpdate = data.lastUpdate
        self.interpolationStartTime = love.timer.getTime()
        self.startX = self.x
        self.startY = self.y
        if data.state ~= self.state then self.currentFrame = 1 end
    end

    local elapsedTime = love.timer.getTime() - self.interpolationStartTime
    if elapsedTime < 0.3 then
        local lerpFactor = elapsedTime / 0.3
        self.x = self.startX + (self.targetX - self.startX) * lerpFactor
        self.y = self.startY + (self.targetY - self.startY) * lerpFactor
    else
        self.x = self.targetX
        self.y = self.targetY
    end

    if data.state == 'die' then
        self.direction = 's'
        self.facing = 's'
    else
        self.direction = data.direction or 's'
        self.facing = data.facing or 's'
    end
    self.state = data.state

    self.animationTime = self.animationTime + dt
    if self.animationTime > 0.2 then
        if self.state == 'die' and self.currentFrame ==
            self.animations.die.frameCount - 1 then
            self.currentFrame = self.animations.die.frameCount - 1
        else
            local animation = self.animations[self.state]
            self.currentFrame = self.currentFrame % animation.frameCount + 1
            self.animationTime = self.animationTime - 0.2
        end
    end
    if self.isFlashing then
        self.isVisible = not self.isVisible
    else
        self.isVisible = true
    end
end

function OtherPlayer:draw()

    local directionIndex
    if self.direction == 's' then
        directionIndex = 0
    elseif self.direction == 'n' then
        directionIndex = 2
    elseif self.direction == 'e' then
        directionIndex = self.facing == 's' and 0 or 2
    elseif self.direction == 'w' then
        directionIndex = self.facing == 's' and 1 or 3
    end

    local animation = self.animations[self.state]
    local frame = animation.frames[directionIndex][self.currentFrame %
                      animation.frameCount + 1]
    local frameWidth, frameHeight = animation.image:getWidth() /
                                        animation.frameCount,
                                    animation.image:getHeight() /
                                        animation.directionCount
    local scaledWidth, scaledHeight = frameWidth * self.scale,
                                      frameHeight * self.scale
    local offsetX, offsetY = scaledWidth / 2, scaledHeight / 2
    local drawX = self.x - offsetX
    local drawY = self.y - offsetY
    if self.isVisible then
        drawManager.bufferDrawEntity(drawX, drawY + (2 * offsetY) -
                                         (self.spriteBottomBuffer * self.scale),
                                     function()
            love.graphics.draw(animation.image, frame, drawX, drawY, 0,
                               self.scale, self.scale)
        end)
    end

    -- draw display name
    local textWidth = pixelFont:getWidth(self.displayName)
    local textHeight = pixelFont:getHeight()
    local textX = self.x - textWidth / 2
    local textY = drawY + textHeight + 15
    drawManager.bufferDrawEntity(0, 0, function()
        love.graphics.setFont(pixelFont)
        love.graphics.setColor(0, 0, 0)
        love.graphics.print(self.displayName, textX + 2, textY + 2)
        love.graphics.setColor(1, 1, 1)
        love.graphics.print(self.displayName, textX, textY)
    end, 1)
end

function OtherPlayer:getBoundingBox() end

return OtherPlayer
