local miniMap = require "miniMap"
local player = require "player"
local Spawner = require "entities.Spawner"
local sdk = require "sdk.sdk"

local hud = {imageFont = nil}

local fullHeart = love.graphics.newImage("img/Interface/heart.png")
local halfHeart = love.graphics.newImage("img/Interface/heart-half.png")

local function formatNumberWithCommas(number)
    return tostring(number):reverse():gsub("(%d%d%d)", "%1,"):reverse():gsub(
               "^,", "")
end

function hud.load()
    hud.imageFont = love.graphics.newImageFont("font/imagefont.png",
                                               " abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.,!?-+/():;%&`'*#=[]\"")
    hud.restartButton = {x = 0, y = 0}
    hud.quitButton = {x = 0, y = 0}
    hud.buttonHeight = 20
    hud.buttonWidth = 100
    hud.textScale = isMobile and 2 or 3
end

function hud.draw()
    local isMobile = love.system.getOS() == "Android" or love.system.getOS() ==
                         "iOS"
    local maxHearts = 5
    local healthPerHeart = 20
    local healthScale = isMobile and 0.06 or 0.1
    local hudX = 20
    local hudY = 20
    local stickOffset = 40
    local scoreMargin = isMobile and 6 or 10
    local shadowOffset = 3

    if player.health <= 0 then
        love.mouse.setVisible(true)
        local gameOverText = "You Died"
        local scoreText = "Score: " .. formatNumberWithCommas(player.score) ..
                              "pts"

        love.graphics.push()
        love.graphics.setFont(hud.imageFont)
        love.graphics.scale(hud.textScale, hud.textScale)

        local screenWidth = love.graphics.getWidth() / hud.textScale
        local screenHeight = love.graphics.getHeight() / hud.textScale

        local font = love.graphics.getFont()
        local textHeight = font:getHeight() / hud.textScale

        local buttonSpacing = 10
        local totalHeight =
            textHeight * 2 + scoreMargin + hud.buttonHeight * 2 + buttonSpacing
        if isMobile then
            totalHeight = totalHeight + (scoreMargin * 3)
        end

        local youDiedTextY = (screenHeight - totalHeight) / 2
        local scoreTextY = youDiedTextY + textHeight + scoreMargin

        hud.restartButton.x = (screenWidth - hud.buttonWidth) / 2
        hud.restartButton.y = scoreTextY + textHeight + (buttonSpacing * 2)
        hud.quitButton.x = (screenWidth - hud.buttonWidth) / 2
        hud.quitButton.y = hud.restartButton.y + hud.buttonHeight +
                               buttonSpacing

        love.graphics.setColor(0, 0, 0, 0.5)
        love.graphics.rectangle('fill', 0, 0, screenWidth * hud.textScale,
                                screenHeight * hud.textScale)
        love.graphics.setColor(1, 1, 1)

        love.graphics.printf(gameOverText, 0, youDiedTextY, screenWidth,
                             'center')
        love.graphics.printf(scoreText, 0, scoreTextY, screenWidth, 'center')

        local buttonTextScale = 0.6

        love.graphics.setColor(0, 1, 0)
        love.graphics.rectangle('fill', (screenWidth - hud.buttonWidth) / 2,
                                hud.restartButton.y, hud.buttonWidth,
                                hud.buttonHeight)
        love.graphics.setColor(1, 1, 1)
        love.graphics.push()
        love.graphics.scale(buttonTextScale, buttonTextScale)
        love.graphics.printf("Restart", (screenWidth - hud.buttonWidth) / 2 /
                                 buttonTextScale, (hud.restartButton.y +
                                 hud.buttonHeight / 4) / buttonTextScale,
                             hud.buttonWidth / buttonTextScale, 'center')
        love.graphics.pop()

        love.graphics.setColor(1, 0, 0)
        love.graphics.rectangle('fill', (screenWidth - hud.buttonWidth) / 2,
                                hud.quitButton.y, hud.buttonWidth,
                                hud.buttonHeight)
        love.graphics.setColor(1, 1, 1)
        love.graphics.push()
        love.graphics.scale(buttonTextScale, buttonTextScale)
        love.graphics.printf("Quit", (screenWidth - hud.buttonWidth) / 2 /
                                 buttonTextScale, (hud.quitButton.y +
                                 hud.buttonHeight / 4) / buttonTextScale,
                             hud.buttonWidth / buttonTextScale, 'center')
        love.graphics.pop()
        love.graphics.pop()

    else

        for i = 1, maxHearts do
            local heartX = (i - 1) * ((fullHeart:getWidth() * healthScale) + 10)
            local heartImage = nil

            if player.health >= i * healthPerHeart then
                heartImage = fullHeart
            elseif player.health > (i - 1) * healthPerHeart then
                heartImage = halfHeart
            end

            if heartImage then
                love.graphics.setColor(0, 0, 0)
                love.graphics.draw(heartImage, heartX + hudX + shadowOffset,
                                   hudY + shadowOffset, 0, healthScale,
                                   healthScale)
                love.graphics.setColor(1, 1, 1)
                love.graphics.draw(heartImage, heartX + hudX, hudY, 0,
                                   healthScale, healthScale)
            end
        end

        local scoreLabelY = hudY + (fullHeart:getHeight() * healthScale) +
                                scoreMargin

        love.graphics.push()
        love.graphics.setFont(hud.imageFont)
        love.graphics.scale(hud.textScale, hud.textScale)

        local scaledHudX = hudX / hud.textScale
        local scaledScoreLabelY = scoreLabelY / hud.textScale

        love.graphics.print(formatNumberWithCommas(player.score) .. 'pts',
                            scaledHudX, scaledScoreLabelY)

        local scaledWaveLabelY = scaledScoreLabelY +
                                     ((love.graphics.getFont():getHeight(
                                         formatNumberWithCommas(player.score)) +
                                         scoreMargin * 3) / hud.textScale)
        love.graphics.print('Wave: ' .. Spawner.wave, scaledHudX,
                            scaledWaveLabelY)

        local scaledNameLabelY = scaledWaveLabelY +
                                     ((love.graphics.getFont():getHeight("Wave") +
                                         scoreMargin * 3) / hud.textScale)
        love.graphics.print(player.displayName, scaledHudX, scaledNameLabelY)

        if Spawner.waveCooldown > 0 and not Spawner.waveReady then
            local nextWaveText = "NEXT WAVE IN"
            local countdownText = tostring(math.ceil(Spawner.waveCooldown)) ..
                                      "s"

            local screenWidth = love.graphics.getWidth() / hud.textScale
            local screenHeight = love.graphics.getHeight() / hud.textScale

            local font = love.graphics.getFont()
            local textHeight = font:getHeight() / hud.textScale

            local textsTotalHeight = textHeight * 2 + scoreMargin
            local nextWaveTextY = (screenHeight - textsTotalHeight) / 2
            local countdownTextY = nextWaveTextY + textHeight + scoreMargin

            love.graphics.printf(nextWaveText, 0, nextWaveTextY, screenWidth,
                                 'center')
            love.graphics.printf(countdownText, 0, countdownTextY, screenWidth,
                                 'center')

        end

        if isMobile then
            if not player.leftThumbstick.active then
                love.graphics.origin()
                love.graphics.setColor(0.75, 0.75, 0.75, 0.5)
                love.graphics.circle("fill", stickOffset +
                                         player.leftThumbstick.outerRadius,
                                     love.graphics.getHeight() - stickOffset -
                                         player.leftThumbstick.outerRadius,
                                     player.leftThumbstick.outerRadius)
                love.graphics.setColor(1, 1, 1)
            else
                love.graphics.origin()
                love.graphics.setColor(0.75, 0.75, 0.75, 0.5)
                love.graphics.circle("fill", player.leftThumbstick.baseX,
                                     player.leftThumbstick.baseY,
                                     player.leftThumbstick.outerRadius)

                love.graphics.setColor(1, 1, 1)
                love.graphics.circle("fill", player.leftThumbstick.innerX,
                                     player.leftThumbstick.innerY,
                                     player.leftThumbstick.innerRadius)
            end
            if not player.rightThumbstick.active then
                love.graphics.origin()
                love.graphics.setColor(0.75, 0.75, 0.75, 0.5)
                love.graphics.circle("fill", love.graphics.getWidth() -
                                         stickOffset -
                                         player.rightThumbstick.outerRadius,
                                     love.graphics.getHeight() - stickOffset -
                                         player.rightThumbstick.outerRadius,
                                     player.rightThumbstick.outerRadius)
                love.graphics.setColor(1, 1, 1)
            else
                love.graphics.origin()
                love.graphics.setColor(0.75, 0.75, 0.75, 0.5)
                love.graphics.circle("fill", player.rightThumbstick.baseX,
                                     player.rightThumbstick.baseY,
                                     player.rightThumbstick.outerRadius)

                love.graphics.setColor(1, 1, 1)
                love.graphics.circle("fill", player.rightThumbstick.innerX,
                                     player.rightThumbstick.innerY,
                                     player.rightThumbstick.innerRadius)
            end
        end

        love.graphics.pop()
        miniMap.draw()
    end
end

function hud.mousepressed(x, y, button)
    if button == 1 and player.health <= 0 then
        local scaledX = x / hud.textScale
        local scaledY = y / hud.textScale

        if scaledX >= hud.restartButton.x and scaledX <=
            (hud.restartButton.x + hud.buttonWidth) and scaledY >=
            hud.restartButton.y and scaledY <=
            (hud.restartButton.y + hud.buttonHeight) then
            sdk.message({restart = true})
            player.health = 100
            player.score = 0
            player.state = 'idle'
        end
        if scaledX >= hud.quitButton.x and scaledX <=
            (hud.quitButton.x + hud.buttonWidth) and scaledY >= hud.quitButton.y and
            scaledY <= (hud.quitButton.y + hud.buttonHeight) then
            sdk.quit()
        end
    end
end

return hud
