local sdk = require "sdk.sdk"
local camera = require "camera"
local config = require "config"
local drawManager = require "drawManager"

local Player = {}
Player.__index = Player

function Player:new()
    local self = setmetatable({}, Player)
    self.x = 160
    self.y = 160
    self.scale = 6
    self.speed = 200
    self.direction = 's'
    self.facing = 's'
    self.currentFrame = 1
    self.animationTime = 0
    self.attackDuration = 0.8
    self.animations = {}
    self.state = 'idle'
    self.spriteBottomBuffer = 13 -- annoying buffer of transparent pixels
    self.crosshairImage = nil
    self.crosshairQuad = nil
    self.crosshairSize = 10
    self.msgTime = 0
    self.health = 100
    self.maxHealth = 100
    self.score = 0
    self.moveX = 0
    self.moveY = 0
    self.leftThumbstick = {
        active = false,
        baseX = 0,
        baseY = 0,
        innerX = 0,
        innerY = 0,
        outerRadius = 60,
        innerRadius = 30,
        touchId = 0,
        touchStartX = 0,
        touchStartY = 0
    }
    self.rightThumbstick = {
        active = false,
        baseX = 0,
        baseY = 0,
        currentX = 0,
        currentY = 0,
        outerRadius = 60,
        innerRadius = 30
    }
    self.angle = 0
    self.displayName = ""
    self.knockback = false
    self.knockbackDuration = 0.5
    self.knockbackStartTime = 0
    self.knockbackSpeed = 300
    self.knockbackX = 0
    self.knockbackY = 0
    self.isFlashing = false
    self.flashDuration = 0.5
    self.flashTimer = 0
    self.isVisible = true
    return self
end

function Player:loadAnimation(imagePath, frameCount, directionCount)
    local image = love.graphics.newImage(imagePath)
    local frames = {}
    local imageWidth = image:getWidth()
    local imageHeight = image:getHeight()

    for i = 0, directionCount - 1 do
        frames[i] = {}
        for j = 0, frameCount - 1 do
            frames[i][j + 1] = love.graphics.newQuad(
                                   j * imageWidth / frameCount,
                                   i * imageHeight / directionCount,
                                   imageWidth / frameCount,
                                   imageHeight / directionCount, imageWidth,
                                   imageHeight)
        end
    end

    return {
        image = image,
        frames = frames,
        frameCount = frameCount,
        directionCount = directionCount
    }
end

function Player:load()
    self.animations.idle = self:loadAnimation('img/Human/Idle.png', 16, 4)
    self.animations.walk = self:loadAnimation('img/Human/Walk.png', 4, 4)
    self.animations.attack = self:loadAnimation('img/Human/Attack.png', 4, 4)
    self.animations.damage = self:loadAnimation('img/Human/Dmg.png', 4, 4)
    self.animations.die = self:loadAnimation('img/Human/SoulDie.png', 12, 1)
    self.crosshairImage = love.graphics.newImage("img/Interface/crosshair.png")
    self.crosshairQuad = love.graphics.newQuad(3, 1, 10, 10,
                                               self.crosshairImage:getDimensions())
    self.x = 160
    self.y = 160
    self.currentFrame = 1
    self.animationTime = 0
    self.state = 'idle'
    self.health = 100
    self.score = 0
    self.moveX = 0
    self.moveY = 0
    self.angle = 0
    self.direction = 's'
    self.facing = 's'
    self.displayName = ""
    self.knockback = false
    self.knockbackDuration = 0.5
    self.knockbackStartTime = 0
    self.knockbackSpeed = 300
    self.knockbackX = 0
    self.knockbackY = 0
    self.isFlashing = false
    self.flashDuration = 0.7
    self.flashTimer = 0
    self.isVisible = true
end

function Player:getWorldMousePosition()
    local screenX, screenY = love.mouse.getPosition()
    local worldX = screenX + camera.x
    local worldY = screenY + camera.y
    return worldX, worldY
end

function Player:attack()
    if self.state == 'die' then return end
    if (self.state ~= 'attack' and self.state ~= 'damage' and not self.knockback and
        self.health > 0) then
        self.currentFrame = 1
        self.animationTime = 0
        self.state = 'attack'
    end
end

function Player:keypressed(key) if key == 'space' then self:attack() end end

function Player:mousepressed(x, y, button)
    if love.system.getOS() ~= "Android" and love.system.getOS() ~= "iOS" then
        if button == 1 then self:attack() end
    end
end

function Player:touchpressed(id, x, y, dx, dy, pressure)
    if x < love.graphics.getWidth() / 2 then
        self.leftThumbstick.active = true
        self.leftThumbstick.touchId = id
        self.leftThumbstick.touchStartX = x
        self.leftThumbstick.touchStartY = y
        self.leftThumbstick.baseX, self.leftThumbstick.baseY = x, y
        self.leftThumbstick.innerX, self.leftThumbstick.innerY = x, y
    elseif x > love.graphics.getWidth() / 2 then
        self.rightThumbstick.active = true
        self.rightThumbstick.touchId = id
        self.rightThumbstick.touchStartX = x
        self.rightThumbstick.touchStartY = y
        self.rightThumbstick.baseX, self.rightThumbstick.baseY = x, y
        self.rightThumbstick.innerX, self.rightThumbstick.innerY = x, y
    end
end

function Player:touchmoved(id, x, y, dx, dy, pressure)
    if self.leftThumbstick.touchId and id == self.leftThumbstick.touchId then
        local dirX, dirY = x - self.leftThumbstick.touchStartX,
                           y - self.leftThumbstick.touchStartY
        local distance = math.sqrt(dirX ^ 2 + dirY ^ 2)
        local maxDistance = self.leftThumbstick.outerRadius

        local adjustedDistance = (distance / maxDistance) ^ 0.5
        local speedScale = math.min(adjustedDistance, 1)

        self.moveX, self.moveY = self:normalize(dirX, dirY)

        self.moveX = self.moveX * speedScale
        self.moveY = self.moveY * speedScale

        if distance > maxDistance then
            distance = maxDistance
            local angle = math.atan2(dirY, dirX)
            self.leftThumbstick.innerX =
                self.leftThumbstick.baseX + maxDistance * math.cos(angle)
            self.leftThumbstick.innerY =
                self.leftThumbstick.baseY + maxDistance * math.sin(angle)
        else
            self.leftThumbstick.innerX, self.leftThumbstick.innerY = x, y
        end
    elseif self.rightThumbstick.touchId and id == self.rightThumbstick.touchId then
        local dirX, dirY = x - self.rightThumbstick.touchStartX,
                           y - self.rightThumbstick.touchStartY
        local distance = math.sqrt(dirX ^ 2 + dirY ^ 2)
        local maxDistance = self.rightThumbstick.outerRadius

        local adjustedDistance = (distance / maxDistance) ^ 0.5

        if distance > maxDistance then
            distance = maxDistance
            local angle = math.atan2(dirY, dirX)
            self.rightThumbstick.innerX =
                self.rightThumbstick.baseX + maxDistance * math.cos(angle)
            self.rightThumbstick.innerY =
                self.rightThumbstick.baseY + maxDistance * math.sin(angle)
        else
            self.rightThumbstick.innerX, self.rightThumbstick.innerY = x, y
        end
    end
end

function Player:touchreleased(id, x, y, dx, dy, pressure)
    if self.leftThumbstick.touchId and id == self.leftThumbstick.touchId then
        self.leftThumbstick.touchId = nil
        self.moveX, self.moveY = 0, 0
        self.leftThumbstick.active = false
    elseif self.rightThumbstick.touchId and id == self.rightThumbstick.touchId then
        self.rightThumbstick.touchId = nil
        self.rightThumbstick.active = false
    end
end

function Player:normalize(x, y)
    local length = math.sqrt(x * x + y * y)
    if length == 0 then
        return 0, 0
    else
        return x / length, y / length
    end
end

function Player:setDisplayName(displayName) self.displayName = displayName end

function Player:setScore(score)
    if self.state ~= 'die' then self.score = score end
end

function Player:update(dt)
    self.animationTime = self.animationTime + dt
    if self.animationTime > 0.2 then
        if self.state == 'die' and self.currentFrame ==
            self.animations.die.frameCount - 1 then
            self.currentFrame = self.animations.die.frameCount - 1
        else
            if (self.state == 'attack' or self.state == 'damage') and
                self.currentFrame == self.animations.attack.frameCount then
                if self.state == 'damage' then
                    self.isFlashing = true
                    self.flashTimer = 0
                end
                self.state = 'idle'
            end
            if self.rightThumbstick.active then self.state = 'attack' end
            local animation = self.animations[self.state]
            self.currentFrame = self.currentFrame % animation.frameCount + 1
            self.animationTime = self.animationTime - 0.2
        end
    end

    if self.state ~= 'damage' and self.state ~= 'die' then
        local isMovingUp = love.keyboard.isDown("up") or
                               love.keyboard.isDown("w")
        local isMovingDown = love.keyboard.isDown("down") or
                                 love.keyboard.isDown("s")
        local isMovingLeft = love.keyboard.isDown("left") or
                                 love.keyboard.isDown("a")
        local isMovingRight = love.keyboard.isDown("right") or
                                  love.keyboard.isDown("d")

        if self.state ~= 'damage' and self.state ~= 'attack' then
            local isMoving = isMovingUp or isMovingDown or isMovingLeft or
                                 isMovingRight or self.moveX ~= 0 or self.moveY ~=
                                 0
            self.state = isMoving and 'walk' or 'idle'
        end

        local moveX, moveY = 0, 0
        if isMovingUp then
            moveY = -1
        elseif isMovingDown then
            moveY = 1
        end

        if isMovingLeft then
            moveX = -1
        elseif isMovingRight then
            moveX = 1
        end

        if love.system.getOS() == "Android" or love.system.getOS() == "iOS" then
            self.x = self.x + self.speed * self.moveX * dt
            self.y = self.y + self.speed * self.moveY * dt
        else
            self.x = self.x + self.speed * moveX * dt
            self.y = self.y + self.speed * moveY * dt
        end

        local angle = self.angle
        if love.system.getOS() == "Android" or love.system.getOS() == "iOS" then
            if self.rightThumbstick.active then
                angle = math.atan2(self.rightThumbstick.innerY -
                                       self.rightThumbstick.baseY,
                                   self.rightThumbstick.innerX -
                                       self.rightThumbstick.baseX)
            end
        else
            local mouseX, mouseY = self:getWorldMousePosition()
            angle = math.atan2(mouseY - self.y, mouseX - self.x)
        end

        if angle < 0 then angle = angle + 2 * math.pi end

        if (angle > 3 * math.pi / 2) and (angle <= math.pi * 2) then
            self.direction, self.facing = 'n', 'n'
        elseif (angle >= 0) and (angle <= math.pi / 2) then
            self.direction, self.facing = 's', 's'
        elseif (angle > math.pi / 2) and (angle <= math.pi) then
            self.direction, self.facing = 'w', 's'
        elseif (angle > math.pi) then
            self.direction, self.facing = 'w', 'w'
        end

        self.angle = angle
    end
    if self.knockback then
        local elapsedTime = love.timer.getTime() - self.knockbackStartTime
        if elapsedTime < self.knockbackDuration then
            local t = elapsedTime / self.knockbackDuration
            local speedFactor = (1 - t * t) * self.knockbackSpeed
            self.x = self.x + self.knockbackX * speedFactor * dt
            self.y = self.y + self.knockbackY * speedFactor * dt
        else
            self.knockback = false
        end
    end
    if self.state ~= 'die' then

        self.msgTime = self.msgTime + dt
        if self.msgTime > 0.1 then
            self.msgTime = 0
            sdk.message({
                x = self.x * config.worldScale,
                y = self.y * config.worldScale,
                direction = self.direction,
                facing = self.facing,
                state = self.state,
                isFlashing = self.isFlashing
            })
        end
    end
    if self.isFlashing then
        self.flashTimer = self.flashTimer + dt
        if self.flashTimer >= self.flashDuration then
            self.isFlashing = false
            self.isVisible = true
        else
            self.isVisible = not self.isVisible
        end
    end
end

function Player:draw()
    if self.isVisible then
        local directionIndex
        if self.state == 'die' then
            directionIndex = 0
        elseif self.direction == 's' then
            directionIndex = 0
        elseif self.direction == 'n' then
            directionIndex = 2
        elseif self.direction == 'e' then
            directionIndex = self.facing == 's' and 0 or 2
        elseif self.direction == 'w' then
            directionIndex = self.facing == 's' and 1 or 3
        end

        local animation = self.animations[self.state]
        local frame = animation.frames[directionIndex][self.currentFrame %
                          animation.frameCount + 1]
        local frameWidth, frameHeight = animation.image:getWidth() /
                                            animation.frameCount,
                                        animation.image:getHeight() /
                                            animation.directionCount
        local scaledWidth, scaledHeight = frameWidth * self.scale,
                                          frameHeight * self.scale
        local offsetX, offsetY = scaledWidth / 2, scaledHeight / 2
        local drawX = self.x - offsetX
        local drawY = self.y - offsetY
        drawManager.bufferDrawEntity(drawX, drawY + (2 * offsetY) -
                                         (self.spriteBottomBuffer * self.scale),
                                     function()
            love.graphics.draw(animation.image, frame, drawX, drawY, 0,
                               self.scale, self.scale)
        end)
    end

    if self.state ~= 'die' and
        (love.system.getOS() ~= "Android" and love.system.getOS() ~= "iOS") then
        local mouseX, mouseY = self:getWorldMousePosition()
        local crosshairScaledSize = self.crosshairSize * self.scale
        local crosshairDrawX = mouseX - crosshairScaledSize / 2
        local crosshairDrawY = mouseY - crosshairScaledSize / 2
        drawManager.bufferDrawEntity(crosshairDrawX, crosshairDrawY, function()
            love.graphics.draw(self.crosshairImage, self.crosshairQuad,
                               crosshairDrawX, crosshairDrawY, 0, self.scale,
                               self.scale)
        end, 1)
    end
end

function Player:getBoundingBox()
    if self.state ~= 'attack' then return 0, 0, 0, 0 end
    local frameWidth, frameHeight, x, y = 0, 0, 0, 0

    if self.currentFrame == 1 then
        frameWidth, frameHeight = 10, 8
    elseif self.currentFrame == 2 then
        frameWidth, frameHeight = 14, 14
    elseif self.currentFrame == 3 then
        frameWidth, frameHeight = 8, 10
    end
    local scaledWidth, scaledHeight = frameWidth * self.scale,
                                      frameHeight * self.scale
    if self.facing == 'w' then
        x = self.x - scaledWidth
        y = self.y - scaledHeight + 30
    elseif self.facing == 'n' then
        x = self.x
        y = self.y - scaledHeight + 30
    elseif self.facing == 's' and self.direction == 's' then
        x = self.x
        y = self.y - 30
    else
        x = self.x - scaledWidth
        y = self.y - 30
    end

    return x, y, scaledWidth, scaledHeight
end

function Player:getBoundingHurtBox()
    local frameWidth, frameHeight = 8, 8
    local scaledWidth, scaledHeight = frameWidth * self.scale,
                                      frameHeight * self.scale

    return self.x - scaledWidth / 2, self.y - scaledHeight / 2, scaledWidth,
           scaledHeight
end

function Player:onCollision(damage, entity)
    if self.state ~= 'damage' and self.state ~= 'die' and not self.isFlashing then
        self.health = self.health - damage
        if self.health > 0 then
            sdk.message({damage = true, health = self.health})
            self.state = 'damage'
            self.currentFrame = 1
            -- knockback
            local dx = self.x - entity.x
            local dy = self.y - entity.y
            local mag = math.sqrt(dx * dx + dy * dy)
            self.knockbackX = dx / mag
            self.knockbackY = dy / mag
            self.knockback = true
            self.knockbackStartTime = love.timer.getTime()
        else
            sdk.message({die = true})
            self.state = 'die'
            self.currentFrame = 1
        end
    end
end

local playerInstance = Player:new()

setmetatable(playerInstance, {__index = function(t, k) return Player[k] end})

return playerInstance
